﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.Xml;
using System.Threading;
using HeuristicLab.Operators.MPISupport;
using HeuristicLab.Core;

namespace HeuristicLab.MPIAlgorithmRunner {
  class Program {
    EventWaitHandle waitHandle = new EventWaitHandle(false, EventResetMode.ManualReset);

    static void Main(string[] args) {
      if (args.Length != 3) {
        Console.WriteLine("Args:" + args.Length);

        throw new ArgumentException("You must provide 3 arguments - the algorithm file, the result file and the update interval (in ms)");
      }

      string fileName = args[0];
      string resultName = args[1];
      int updateInterval = int.Parse(args[2]);

      using (new MPI.Environment(ref args)) {
        MPI.Communicator communicator = MPI.Communicator.world.Clone() as MPI.Communicator;

        if (communicator.Rank == 0) {
          IAlgorithm alg = XmlParser.Deserialize<IAlgorithm>(fileName);
          MPITransportWrapper<IAlgorithm> transport = new MPITransportWrapper<IAlgorithm>(alg);

          int clients = communicator.Size - 1;
          Console.WriteLine("Sending alg to " + clients + " clients");
          for (int i = 0; i < clients; i++) {
            int client = i + 1;
            communicator.Send(transport, client, 0);
          }

          ItemList<ResultCollection> results = new ItemList<ResultCollection>();
          
          bool[] finished = new bool[clients];
          int finishedCount = 0;

          while (finishedCount != clients) {
            for (int i = 0; i < clients; i++) {
              if (!finished[i]) {
                int client = i + 1;
                ResultCollection result = communicator.Receive<MPITransportWrapper<ResultCollection>>(client, 1).InnerItem;

                Console.WriteLine("Received result " + result);

                if (results.Count != clients) {
                  results.Add(result);
                } else {
                  results[i] = result;
                }

                Console.WriteLine("Probing...");
                if (communicator.ImmediateProbe(client, 2) != null) {
                  finished[i] = true;
                  finishedCount++;
                }
              }
            }

            Console.WriteLine("Update results");
            lock (resultName) {
              XmlGenerator.Serialize(results, resultName);
            }
          }

          Console.WriteLine("Finished.");
        } else {
          Program p = new Program();
          p.StartAlgorithm(updateInterval, communicator);
        }
      }
    }

    public void StartAlgorithm(int updateInterval, MPI.Communicator communicator) {
      IAlgorithm alg = communicator.Receive<MPITransportWrapper<IAlgorithm>>(0, 0).InnerItem;

      Console.WriteLine("Starting algorithm...");

      alg.Stopped += new EventHandler(algorithm_Stopped);
      waitHandle.Reset();

      alg.Prepare(true);
      alg.Start();

      Timer t = new Timer(delegate(object state) {
        if (alg.ExecutionState == ExecutionState.Started) {
          Console.WriteLine("Pausing alg...");
          alg.Pause();

          while (alg.ExecutionState != ExecutionState.Paused) {
            Thread.Sleep(100);
          }

          communicator.Send<MPITransportWrapper<ResultCollection>>(
            new MPITransportWrapper<ResultCollection>(alg.Results), 0, 1);

          Console.WriteLine("Sending update...");

          Console.WriteLine("Resuming alg...");
          alg.Start();
        }
      }, null, updateInterval, updateInterval);

      waitHandle.WaitOne();

      communicator.Send<int>(communicator.Rank, 0, 2);
      communicator.Send<MPITransportWrapper<ResultCollection>>(
          new MPITransportWrapper<ResultCollection>(alg.Results), 0, 1);
    }

    void algorithm_Stopped(object sender, EventArgs e) {
      waitHandle.Set();
    }
  }
}
